#!/usr/bin/perl -w

# Just In Time Authentication Mechanism
#######################################

# Version Information
#####################
# Version Number 1.0.0
# Last Updated: 1-7-2004 by Phill Berrie.

# Copyright Information
#######################
# This script was developed by Phill Berrie for the Authenticated Electronic Editions and
# Electronic Collaboration in Humanities Research Projects of the Australian Scholarly 
# Editions Centre at UNSW at ADFA in Canberra Australia. It is freely available for non-
# commercial and educational use and repurposing under the licencing restrictions listed 
# below. The licence is based on the BSD Open Source licence template.
#
# Copyright (c) The University of New South Wales 2004
# All rights reserved
# Inquiries should be made to Australian Scholarly Editions Centre, School of Humanities
# and Social Sciences, University of New South Wales at ADFA, Canberra ACT 2600 AUSTRALIA.
#
# Redistribution and use in source and binary forms, with or without modification, for
# non-commercial use are permitted provided that the following conditions are met:
#
#	 	Redistributions of source code must retain the above copyright notice, this list 
#     of conditions and the following disclaimer.
#	 	Redistributions in binary form must reproduce the above copyright notice, this 
#     list of conditions and the following disclaimer in the documentation and/or other 
#     materials provided with the distribution.
#	 	Neither the names of the Authenticated Electronic Editions Project and the Electronic
#     Collaboration in Humanities Research Project nor the names of its contributors may 
#     be used to endorse or promote products derived from this software without specific 
#     prior written permission.
#
# TO THE EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT 
# HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT 
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.

# What it does
##############
# This script is used for authenticating the textual content of SGML-like elements
# in files that are being edited with regard to their markup. The use of this script
# will reduce the proofreading overhead in applications where the accuracy of transcription
# is paramount.
#
# Correct use of this script requires a preparatory scan of the file it is meant to 
# authenticate. This scan produces a <filename>.keys file, which is used to authenticate
# the file after an editing session. The second time the file is scanned the textual content
# of its new state is compared to the saved version and any authentication failures are
# recorded in the .keys file to assist in the correction of the corruption. More details
# on how to use this utility is provided in the accompanying documentation.

# System Requirements
#####################
# This script is written in Perl and requires the use of Perl 5.8.1 or later to gain
# access to the Unicode implementation that was built into the language at this point.
# The design of the script is to be platform agnostic and should produce the same results 
# on platforms with different underlying character representations and end-of-line 
# characteristics. This capability has currently been tested on Mac OS X, Windows 2000 and XP, 
# Linux and Solaris and applies to the following character representations: ASCII, ISO-646, 
# ISO-10646, utf-8, utf-16 either big-endian or little-endian with and without 
# byte-order-marks.
#
# The correct use of this utility requires the inclusion of an id attribute into the 
# text elements of interest. It is these elements that will be authenticated by the utility.

use Encode ;
use Encode::Guess ;

my $hyphenMode = "Include" ;                # Setting the default mode for authentication.
my $fileName = $ARGV[0] ;                   # Name of file to be read in by the script.
if ( ( @ARGV > 1 ) and ( $ARGV[1] eq "--ignore") ) 
{ 
  $hyphenMode = "Ignore" ;                  # Mode used for Just In Time Markup authentication.
}

# Declaring working variables and constants.
my $startMarkup = "<" ;                     # Assume the standard characters are used
my $endMarkup = ">" ;                       # for indicating start and end of markup.

my $fileID = $fileName ;                    # Calculate the fileID string from the target
                                            # file name for internal identification of keys.
$fileID =~ s/\.tspt// ;                     # For JITM transcription files we remove ".tspt".

my $keyFileName = "" ;                      # Variable for holding calculated key file name.

my @keyDataArray = "" ;                     # Array for authentication filekeys.
my $keyDataScalar = "" ;                    # Scalar equivalent of above.

my @fileDataArray = "" ;                    # Array for holding input and text elements.
my $fileDataScalar = "" ;                   # Scalar equivalent of above.

my $fileDataCheckSum ;                      # Check sum of file data as scalar.

my $fileKey = 0 ;                           # Checksum of the content of the data file.
my $teID = "" ;                             # Variable to hold text element ID.
my $keyCheckVar = "" ;                      # Variable for checking filekey data.
my $nameString = "" ;                       # variable for holding element name string.
my $arrayElement = "" ;                     # Reusable working variable for doing array processing.


# Determine the filekey authentication file name.
$keyFileName = calculateKeyFileName( $fileName ) ;

# Read in the authentication key file if it exists and reset it to the default state.
# Leave the resulting information in two forms for ease of calculation (i.e. an array
# for processing incrementally and a string variable for searching.
if ( -e $keyFileName )
{
  @keyDataArray = readDecodeFile( $keyFileName ) ;
	$keyDataScalar = join( "\n", @keyDataArray ) ;
	$keyDataScalar =~ s/\n\n/\n/g ;
	$keyDataScalar =~ s/:AUTHENTICATION FAILED//g ;
	$keyDataScalar =~ s/:TEXT ELEMENT NOT FOUND//g ;
	$keyDataScalar =~ s/:TEXT ELEMENT OUT OF ORDER//g ;
	@keyDataArray = split(/\n/, $keyDataScalar) ;
}


# Read in the data file into an array using the 'Encode' library to do the character set 
# identification and conversion to Perl's internal utf-8 representation.
@fileDataArray = readDecodeFile( $fileName ) ;
$fileDataScalar = join( "\n", @fileDataArray ) ;

# First step in preparing the file data for authentication is to delete all soft hyphen
# entity references if we are in the 'ignore' mode. This is a Just In Time Markup requirement
# and is not the default. 
if ( $hyphenMode eq "Ignore" ) {  $fileDataScalar =~ s/&shy;//g ; }

# First stage of authentication involves check-summing the entire file. Failing this check
# sum does not necessarily indicate a problem with the file.
$fileDataCheckSum = checkSum( $fileDataScalar ) ;

# In the next step we change the structure of the file data array so that the content can
# be separated from markup. The resulting array has content and markup as different elements
# in the array, but in the original order as the file that was read in. This shows the 
# dichotomous nature of the data and allows us to restructure the data in a way that better
# suits the task at hand. 
$fileDataScalar =~ s/\n//g ;
$fileDataScalar =~ s/$startMarkup/\n$startMarkup/g ;
$fileDataScalar =~ s/$endMarkup/$endMarkup\n/g ;
$fileDataScalar =~ s/\n\n/\n/g ;
$fileDataScalar = substr($fileDataScalar, 1, length( $fileDataScalar ) ) ;

# Collate array elements into meaningful sections based on those elements with
# an id attribute then delete all array elements that do not have an id attribute.
#
# For example:
# 
#   <tag id="XXX">
#   This is the tags content
#   </tag>
#
# becomes
# 
#   <tag id="XXX">This is the tags content</tag>
#
# Currently this assumes a simple structure with no embedded tags having an id attribute.
@fileDataArray = "" ;
my $counter = 0 ;
foreach $arrayElement ( split( /\n/, $fileDataScalar ) )
{
  if ( ( $arrayElement =~ / id=\".*"/ ) and ( $fileDataArray[ $counter ] ne "" ) )
  {
    $fileDataArray[ ++$counter ] = $arrayElement ;
  }
  else
  {
    $fileDataArray[ $counter ] = $fileDataArray[ $counter ] . $arrayElement ;
  }
}

# There may be array elements that are elements without an id attribute. These will now be
# removed.
@fileDataArray = grep( / id=\".*"/, @fileDataArray ) ;

# Similarly the resulting array elements may have some trailing data outside the scope of 
# the elements with id attributes. These need to be trimmed.
$counter = 0 ;
my $finishTag = "" ;
foreach $arrayElement ( @fileDataArray )
{
  $finishTag = $arrayElement ;
  $finishTag =~ s/$startMarkup(\S*).*|$startMarkup(\S*)$endMarkup.*/$startMarkup\/$1$endMarkup/ ;
  $arrayElement =~ s/(.*)$finishTag.*/$1$finishTag/ ;
}

# Now we will iterate through the array and generate the fileKey checksum for all the 
# text elements and print them to the key file, creating it if necessary or overwriting
# if it already exists.
open ( OUTPUTFILE, ">", $keyFileName ) ;
	if ( $keyDataScalar eq "" )
	{
	  # File does not exist or is empty therefore we are going to create it and
	  # return a message to this effect to standard output.
	  print "JITAM: Key File Created. \n" ;
	 
	  # Key file is empty or does not exist. Output results to create this file.
		print OUTPUTFILE $fileName . ":" . $fileDataCheckSum . "\n" ;

    # This section simply outputs the element type, id attribute value and checksum for
    # each text element in the file in the order, which it appears.
		foreach $arrayElement ( @fileDataArray )
		{
		  $nameString = $arrayElement ;
		  $nameString =~ m/ /g ;
		  $nameString = $startMarkup . "/" . substr($nameString, 1, ( pos($nameString) - 2) ) . $endMarkup ;
		  $arrayElement =~ s/(.*)$nameString.*/$1$nameString/ ;
		  $teID = getAttribute( "id", $arrayElement ) ;
		  $arrayElement = separateContent( $arrayElement ) ;
		  $fileKey = checkSum( $arrayElement ) ;
		  $nameString = substr($nameString, 2, (length($nameString) - 3) ) ;
		  $keyCheckVar = $nameString . ":" . $teID . ":" . $fileKey ;
		  print OUTPUTFILE $keyCheckVar . "\n" ;
		}
	}
	else
	{
	  # The key file exists and is going to be annotated with the results of our authentication.
	  
	  # First we check if the calculated checksum for the whole file matches the stored value.
	  # If it does not then we indicate as such, but this is not a strong indication of corruption
	  # as there may have been text modified in elements without an id attribute.
    if ( $keyDataArray[0] !~ ( $fileID . ":" . $fileDataCheckSum ) )
    {
      $keyDataArray[0] = $keyDataArray[0] . ":AUTHENTICATION FAILED" ;
    }
    	
    # Next we scan through the key data to check that the text elements in the file
    # are present and in the right order. This process discovers gross corruptions in 
    # the file.
    my @orderedFileData = @fileDataArray ;
    my $headerData = $keyDataArray[0] ;
    @keyDataArray = @keyDataArray[1..$#keyDataArray] ;
    foreach $arrayElement ( @keyDataArray )
    {
      $teID = $arrayElement ;
      $teID =~ s/.*:(.*):.*/$1/ ;
      chomp($teID) ;
      if ( $fileDataScalar !~ /id="$teID"/ )
      {
        $arrayElement = $arrayElement . ":TEXT ELEMENT NOT FOUND" ;
      }
      else
      {
        if ( $orderedFileData[0] !~ /$teID/ )
        {
          $arrayElement = $arrayElement . ":TEXT ELEMENT OUT OF ORDER" ;
        }
      }
      @orderedFileData = @orderedFileData[1..$#orderedFileData] ;
    }

    # Having possibly modified our key data we have to make a new version of the scalar
    # version of the data to keep them in sync.
	  $keyDataScalar = join( "\n", @keyDataArray ) ;

    # Now we step through each element in our file array and calculate the checksum for
    # each content. Checking to see if the concatenated value of the tag type, id attribute
    # and calculated value of the checksum occurs in the key file data determines whether
    # the key file data is modified to indicate a corruption has occurred.
    foreach $arrayElement ( @fileDataArray )
	  {
	    $nameString = $arrayElement ;
      $nameString =~ m/ /g ;
	    $nameString = $startMarkup . "/" . substr($nameString, 1, ( pos($nameString) - 2) ) . $endMarkup ;
	    $arrayElement =~ s/(.*)$nameString.*/$1$nameString/ ;
	    $teID = getAttribute( "id", $arrayElement ) ;
	    $arrayElement = separateContent( $arrayElement ) ;
	    $fileKey = checkSum( $arrayElement ) ;
	    $nameString = substr($nameString, 2, (length($nameString) - 3) ) ;
	    $keyCheckVar = $nameString . ":" . $teID . ":" . $fileKey . "\n" ;
	    if ( $keyDataScalar !~ $keyCheckVar )
	    {
	    # In this block we are modifying the key file data by appending an authentication failed
	    # message. Note that we are not interested in the calculated value, just the fact that the
	    # comparison failed. The data originally contained in the key file is not changed as it
	    # represents the values of the original file.
	      $keyDataScalar =~ s/$nameString:$teID:(.*)\n/$nameString:$teID:$1:AUTHENTICATION FAILED\n/ ;
	    }
	  }
	 	 
	  print OUTPUTFILE $headerData . "\n" . $keyDataScalar ;
  }
close( OUTPUTFILE) ;

if ( ( $keyDataScalar =~ /FAILED/ ) or  ( $keyDataScalar =~ /TEXT ELEMENT/ ) )
{
  print "JITAM: Authentication Failed. \n" ;
}
else
{
  print "JITAM: Authentication Succeeded. \n" ;
}

######################################################################################
#   Subroutines
######################################################################################

sub getAttribute
# This subroutine assume the input parameter is a string that contains text that follows the
# standard SGML attribute format (i.e id="---"). It returns the content of the quotation
# marks. It will accept either single or double quotation marks assuming double quotes.
{
  my ( $attributeName, $theString ) = @_ ;
  
  if ( $theString =~ m/$attributeName="/ )
  {
	  $theString =~ s/^.+ $attributeName="// ;
	  $theString =~ s/".+// ;
  }
  else
  {
	  $theString =~ s/^.+ $attributeName='// ;
	  $theString =~ s/'.+// ;
  }
  return $theString ;
}

sub checkSum
# Standard JITM check sum algorithm that assumes text is in utf-8 character set.
{
# Initialise our input and output parameters.
  my $theValue = 0 ;
  my ( $contentStream ) = @_ ;
  
# Initialise our encoding independent character value. Rather than depend on the
# value returned by our ordinal value function which may very with platform and
# character set in the future, we determine the local value for the space character
# and determine our own ordinal value as the offset from this value to the character
# value. 
  my $spaceValue = ord( " " ) ;
  
# First thing we need to do is normalise the file data by turning any runs of 
# white space characters into a single space. We shall also chop off any preceding
# or following white space. This complies with the SGML/XML practice of ignoring
# white space in the coding.
  $contentStream =~ s/\s+/ /g ;
  $contentStream =~ s/^ //g ;
  $contentStream =~ s/ $//g ;
  
# Initialisation of debugging variables. These variables and where they are used in 
# redundant code have been indicated with a "##" mark.
  my $stringLength = length( $contentStream ) ; ##
  my $myCharacter = "" ; ##
  my $charValue = 0 ; ##

  for ( my $counter = 0 ; $counter < length( $contentStream ) ; $counter++ )
  {
    $myCharacter = substr( $contentStream, $counter, 1 ) ; ##
    # The internal ordinal value of the character is determined by finding its
    # offset from the platform dependent value. The addition of the value 32 is
    # so that existing legacy key files created using an ASCII dependent algorithm will
    # still authenticate with this character set independent algorithm.
    $charValue = ord( $myCharacter ) - $spaceValue + 32 ; ##
    $theValue = $theValue + ( ( $counter + 1) * $charValue ) ;
  }
  
  return $theValue ;
}

sub separateContent
{
# This subroutine separates out the two different types of information from the
# provided stream of text. It returns the content stream.

  my ( $fileStream ) = @_ ;
	my $tagStream = "" ;                        # Variable for holding tags of file.
	my $textStream = "" ;                       # Variable for holding contents of file.
	my $mode = "TEXT" ;                         # Variable used to indicate type of data.
	my $marker = 0 ;                            # Next position of interest in file data.
  
	my $startMarkup = "<" ;                     # Assume the standard characters are used
	my $endMarkup = ">" ;                       # for indicating start and end of markup.

# This next step is a simple state machine where the text stream is diverted
# into a variable depending on the state of the machine. The state of the machine
# is controlled by the input characters.
while ( $fileStream ne "" )
{
  if ( $mode eq "TEXT" )
  {
    $marker = index( $fileStream, $startMarkup ) ;
    if ( $marker == -1 )
    {
      $textStream = $textStream . $fileStream ;
      $fileStream = "" ;
    }
    else
    {
      $mode = "MARKUP" ;
      if ( $marker != 0 )
      {
	      $textStream = $textStream . substr( $fileStream, 0, $marker ) ;
	      $fileStream = substr( $fileStream, $marker ) ;
      }
    }
  }
  else
  {
    $marker = index( $fileStream, $endMarkup ) ;
    if ( $marker == -1 )
    {
      $tagStream = $tagStream . $fileStream ;
      $fileStream = "" ;
    }
    else
    {
      $mode = "TEXT" ;
      if ( $marker != 0 )
      {
	      $tagStream = $tagStream . substr( $fileStream, 0, ( $marker + 1 ) ) ;
	      $fileStream = substr( $fileStream, ( $marker + 1 ) ) ;
      }
    }
  }
}

  return $textStream ;
}

sub readDecodeFile 
{
	# Read in a file line by line using the 'Encode' library to do the character set identification 
	# and conversion to Perl's internal utf-8 representation. If this is successful return the results 
	# as an array.
	my ( $theFile ) = @_ ;
	my @theFileArray = "" ;
	my $theFileScalar = "" ;
	
	open( INPUTFILE, "<", $theFile ) or die("Can't find file: " . $theFile) ;
	  while ( <INPUTFILE> )
	  {
	    $theFileArray[$#theFileArray++] = decode( "Guess", $_ ) ;
	  }
	  # Trim the last undefined array entry.
	  @theFileArray = @theFileArray[ 0..($#theFileArray - 1) ] ;
	close( INPUTFILE ) ;
	
	# If there are BOMs still included in the file these have to be removed from the first
  # array entry so that the checksum for the entire file will be the same as for versions
  # of the file which do not have BOMs or they have been removed.
  my $BOMCheck = substr( $theFileArray[0], 0, 1 ) ;
  $BOMCheck = ord( $BOMCheck ) ;
  if ( $BOMCheck == 65279 )
  {
    $theFileArray[0] =~ s/.//s ;
  }

  # If reading the file data does not result in a multiple entry array then it is likely
  # that the file does not have linefeed characters to indicate the end of lines. This 
  # probably means that it is a Macintosh file and the problem can be solved by splitting the
  # first member of the array, which contains all the data, into the true array based on
  # the existence of carriage returns (i.e. ASCII 13).
  if ( $theFileArray[0] !~ m/\n/ )
  {
    $theFileScalar = $theFileArray[0] ;
    $theFileScalar =~ s/\r/\n/g ;
  }
  else
  {
    # DOS files also have to have the extraneous return character removed from each array
    # element. Since carriage return are not normally in unix files we will remove all 
    # carriage returns from files that were not originally recognised as Macintosh files.
    $theFileScalar = join( "\n", @theFileArray ) ;
    $theFileScalar =~ s/\r//g ;
    $theFileScalar =~ s/\n\n/\n/g ;
  }

  @theFileArray = split( /\n/, $theFileScalar ) ;
	return @theFileArray ;
}

sub calculateKeyFileName
{
  # This routine calculates a suitable file name for the keys based on the file name provided in the arguments
  # This is only done once, but has been made into a subroutine to tidy up the code of the main program. This 
  # is done by either adding ".keys" to the file name if it does not have a file name extension or replacing
  # the file name extension if it does (eg. "blah.html" creates a new file called "blah.keys")
  my $keyFileName = "" ;
  my ( $fileName ) = @_ ;
  
  if ( $fileName =~ /.keys$/ )
  {
    die( "Illegal: You cannot process a key file" ) ;
  }
  
  if ( $fileName !~ m/\./ ) 
	{
	  $keyFileName = $fileName . ".keys" ;      
	}
	else
	{
	  my @tempFileName = split( /\./, $fileName ) ;
	  $tempFileName[ $#tempFileName ] = "keys" ; 
	  $keyFileName = join( "\.", @tempFileName ) ;
	}
	
	return $keyFileName ;
}